package xmlCsv;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.IOException;
import java.io.UnsupportedEncodingException;
import java.io.Reader;
import java.io.BufferedReader;

import java.util.List;
import java.util.ArrayList;
import java.util.Arrays;

import daruma.util.CharacerCodeConstant;

import xmlCsv.PeekReader;


public class CsvReader
{
	private	static	final	String	SEPARATOR = ",";

	public	CsvReader()
	{
	}

	private	enum TerminatedStatus { TERMINATED_WITH_COMMA ,
					TERMINATED_WITH_NEWLINE ,
					TERMINATED_WITH_EOF ,
					NO_DATA };

	public	List<List<String>>	read( InputStream  in ,
					      String  inputEncoding )
				throws CsvFormatException,
				       UnsupportedEncodingException,
				       IOException
	{
		return( this.read( in , inputEncoding , false ) );
	}

	public	List<List<String>>	read( InputStream  in ,
					      String  inputEncoding ,
					      boolean  debug )
				throws CsvFormatException,
				       UnsupportedEncodingException,
				       IOException
	{
		PeekReader	reader;
		reader = new PeekReader( new InputStreamReader
					 ( in , inputEncoding ) );

		List<List<String>>	lines = new ArrayList<List<String>>();
		int			nextDataLines = 1;

		boolean	continueLine = false;

		for(;;)
		{
			TerminatedStatus	terminated;
			StringBuilder		column = new StringBuilder();

			terminated = this.readColumn( reader ,
						      column ,
						      debug );

			if ( terminated != TerminatedStatus.NO_DATA )
			{
				if ( debug )
				{
					System.err.println( "column = ["
							    + column.toString()
							    + "]"
							    + ": "
							    + terminated );
				}

				if ( ! continueLine )
				{
					lines.add( new ArrayList<String>() );
				}

				if ( terminated == TerminatedStatus
						     .TERMINATED_WITH_NEWLINE )
				{
					continueLine = false;
				}
				else
				{
					continueLine = true;
				}

				lines.get(lines.size() - 1)
					.add( column.toString() );
			}

			if ( terminated == TerminatedStatus.TERMINATED_WITH_EOF
			  || terminated == TerminatedStatus.NO_DATA )
			{
				break;
			}
		}


		if ( debug )
		{
			//
			// debug
			//
			for( int  i = 0  ;  i < lines.size()  ;  i++ )
			{
				System.err.println( "lines[" + i + "]" );

				for( int  j = 0  ;  j < lines.get(i).size()  ;
				     j ++ )
				{
					System.err.println
						( "lines[" + i + "]"
						  + "[" + j + "] = ["
						  + lines.get(i).get(j)
						  + "]" );
				}
			}
		}

		return( lines );
	}


	private	TerminatedStatus	readColumn( PeekReader  reader ,
						    StringBuilder  buf ,
						    boolean  debug )
						throws CsvFormatException ,
						       IOException
	{
		boolean	firstCharacter = true;

		boolean	quoted = false;
		int	quoteCount = 0;

		boolean	finishWithNewline = false;

		int	ch = -1;

		TerminatedStatus	terminatedStatus;

		makeCulumn: for(;;)
		{
			ch = reader.read();

			if ( debug )
			{
				System.err.println( "ch = " + ch + "["
						    + (char)ch + "]" );
			}

			switch( ch )
			{
			case -1: // EOF
				if ( firstCharacter )
				{
					return( TerminatedStatus.NO_DATA );
				}

				if ( quoted && quoteCount != 1 )
				{
					throw new CsvFormatException
						( "unexpected EOF,"
						  + " double quote expected" );
				}

				return( TerminatedStatus
					.TERMINATED_WITH_EOF );


			case ',':
				if ( quoted )
				{
					if ( quoteCount == 1 )
					{
						return( TerminatedStatus
							.TERMINATED_WITH_COMMA );
					}
					else
					{
						buf.append( (char)ch );
					}
				}
				else
				{
					return( TerminatedStatus
						.TERMINATED_WITH_COMMA );
				}

				break;


			case '\r':
			case '\n':
				if ( (quoted && quoteCount == 1)
				   || ! quoted )
				{
					if ( ch == '\r'
					  && reader.peek() == '\n' )
					{
						reader.read();
					}

					return( TerminatedStatus
						.TERMINATED_WITH_NEWLINE );
				}
				else
				{
					buf.append( (char)ch );
				}

				break;


			case '"':
				if ( quoted )
				{
					quoteCount ++;

					if ( quoteCount == 2 )
					{
						buf.append( '"' );
						quoteCount = 0;
					}
				}
				else if ( firstCharacter )
				{
					quoted = true;
				}
				else
				{
					throw new CsvFormatException
					  ( "unquoted double quote found" );
				}

				break;


			default:
				buf.append( (char)ch );
				break;
			}

			firstCharacter = false;
		}
	}


	/*
	// simple implemantation: ignores double quote(") escaping
	public	ArrayList< List<String> >	read( InputStream  in ,
						      String  inputEncoding )
				throws UnsupportedEncodingException ,
				       IOException
	{
		BufferedReader	reader;
		reader = new BufferedReader
			 ( new InputStreamReader( in , inputEncoding ) );

		ArrayList< List<String> >	table;
		table = new ArrayList< List<String> >();

		String  line;
		while( (line = reader.readLine()) != null )
		{
			table.add( Arrays.asList( line.split( SEPARATOR ) ) );
		}

		return( table );
	}
	*/
}
