package daruma.geometry;

import daruma.geometry.Point;
import daruma.geometry.LineString;
import daruma.geometry.Polygon;

import org.w3c.dom.Element;

import daruma.xml.URI;
import daruma.xml.util.ElementUtil;
import daruma.xml.util.XMLParseErrorException;
import daruma.strage_manager.type_definition.ElementName;

import java.util.List;
import java.util.ArrayList;


public class GeometryFormatConverter
{
	private	GeometryFormatConverter()
	{
	}

	public	static	Point	parseGMLPoint( String  coordinatesString )
						    throws GeometryException
	{
		String[]	values = coordinatesString.trim().split( "," );

		if ( values.length < 2 )
		{
			throw new GeometryException
			  ( "expected at least 2 values: "
			    + "\"" + coordinatesString + "\"" );
		}


		double	x;
		try
		{
			x = Double.parseDouble( values[0] );
		}
		catch( NumberFormatException  e )
		{
			throw new GeometryException
			  ( "\"" + values[0] + "\""
			    + " is not a valid number" );
		}

		double	y;
		try
		{
			y = Double.parseDouble( values[1] );
		}
		catch( NumberFormatException  e )
		{
			throw new GeometryException
			  ( "\"" + values[1] + "\""
			    + " is not a valid number" );
		}

		return( new Point( x , y ) );
	}

	public	static	LineString	parseGMLLineString
					  ( String  coordinatesString )
						    throws GeometryException
	{
		LineString	ret = new LineString();

		String[]	points = coordinatesString
						.trim().split( "[ \t\n\r]+" );

		for ( String  pointString : points )
		{
			try
			{
				ret.add( GeometryFormatConverter
					 .parseGMLPoint( pointString ) );
			}
			catch( GeometryException  e )
			{
				throw e;
			}
		}

		return( ret );
	}

	public	static	Polygon	parseGMLPolygon( String  coordinatesString )
						    throws GeometryException
	{
		Polygon	ret = new Polygon();

		String[]	points = coordinatesString
						.trim().split( "[ \t\n\r]+" );

		for ( String  pointString : points )
		{
			try
			{
				ret.add( GeometryFormatConverter
					 .parseGMLPoint( pointString ) );
			}
			catch( GeometryException  e )
			{
				throw e;
			}
		}

		if ( ret.getPoints().size() == 0 )
		{
			throw new GeometryException
				  ( "invalid Polygon format,"
				    + " no points" );
		}

		if ( ! ret.closed() )
		{
			throw new GeometryException
				  ( "invalid Polygon format,"
				    + " Polygon not closed: ["
				    + coordinatesString + "]" );
		}

		return( ret );
	}


	public	static	Polygon	parseGMLBox( String  coordinatesString )
						    throws GeometryException
	{
		String[]	points = coordinatesString
						.trim().split( "[ \t\n\r]+" );

		List<Point>	pointList = new ArrayList<Point>();

		for ( String  pointString : points )
		{
			try
			{
				pointList.add( GeometryFormatConverter
					       .parseGMLPoint( pointString ) );
			}
			catch( GeometryException  e )
			{
				throw e;
			}
		}

		if ( pointList.size() != 2 )
		{
			throw new GeometryException
				  ( "invalid Box format,"
				    + " 2 points expected" );
		}

		Polygon	ret = new Polygon();

		Point	pMin = pointList.get(0);
		Point	pMax = pointList.get(1);

		ret.add( new Point( pMin.getX() , pMin.getY() ) );
		ret.add( new Point( pMax.getX() , pMin.getY() ) );
		ret.add( new Point( pMax.getX() , pMax.getY() ) );
		ret.add( new Point( pMin.getX() , pMax.getY() ) );
		ret.add( new Point( pMin.getX() , pMin.getY() ) );

		return( ret );
	}


	public	static	String	convertGeometryToWKTString( Geometry  geom )
	{
		if ( geom instanceof Point )
		{
			return( GeometryFormatConverter
				.convertPointToWKTString( (Point)geom ) );
		}
		else if ( geom instanceof LineString )
		{
			return( GeometryFormatConverter
				.convertLineStringToWKTString
				 ( (LineString)geom ) );
		}
		else if ( geom instanceof Polygon )
		{
			return( GeometryFormatConverter
				.convertPolygonToWKTString( (Polygon)geom ) );
		}
		else
		{
			return( null );
		}
	}


	public	static	String	convertPointToWKTString( Point  point )
	{
		return( "POINT(" + point.getX() + " " + point.getY() + ")" );
	}

	public	static	String	convertLineStringToWKTString
						( LineString  lineString )
	{
		StringBuffer	buf = new StringBuffer();

		buf.append( "LINESTRING(" );

		boolean	firstPoint = true;
		for ( Point  p : lineString.getPoints() )
		{
			if ( ! firstPoint )
			{
				buf.append( "," );
			}

			buf.append( p.getX() + " " + p.getY() );

			firstPoint = false;
		}

		buf.append( ")" );

		return( buf.toString() );
	}

	public	static	String	convertPolygonToWKTString( Polygon  polygon )
	{
		StringBuffer	buf = new StringBuffer();

		buf.append( "POLYGON((" );

		boolean	firstPoint = true;
		for ( Point  p : polygon.getPoints() )
		{
			if ( ! firstPoint )
			{
				buf.append( "," );
			}

			buf.append( p.getX() + " " + p.getY() );

			firstPoint = false;
		}

		buf.append( "))" );

		return( buf.toString() );
	}


	public	static	Geometry  parseFromGMLElement
					    ( Element  geometryElement )
						    throws GeometryException
	{
		ElementName	geometryTypeElementName
					= new ElementName( geometryElement );

		if ( ! geometryElement.getNamespaceURI().equals( URI.GML ) )
		{
			throw new GeometryException
				  ( "unexpected geometry type \""
				    + geometryTypeElementName + "\"" );
		}


		Element	childElement;

		try
		{
			childElement = ElementUtil.getSingleChildElement
							      ( geometryElement );
		}
		catch( XMLParseErrorException  e )
		{
			throw new GeometryException( e.getMessage() , e );
		}


		Geometry	geom;

		String	coordinatesString;
		try
		{
			if ( geometryTypeElementName.getLocalName()
			     .equals( "Point" ) )
			{
				coordinatesString
				  = GeometryFormatConverter
				    .getCoordinatesString( childElement );

				geom = GeometryFormatConverter
					.parseGMLPoint
					 ( coordinatesString );
			}
			else if ( geometryTypeElementName.getLocalName()
				  .equals( "LineString" ) )
			{
				coordinatesString
				  = GeometryFormatConverter
				    .getCoordinatesString( childElement );

				geom = GeometryFormatConverter
					.parseGMLLineString
					 ( coordinatesString );
			}
			else if ( geometryTypeElementName.getLocalName()
				  .equals( "Polygon" ) )
			{
				coordinatesString
				  = GeometryFormatConverter
				    .getPolygonCoordinatesString
							( childElement );

				geom = GeometryFormatConverter
					.parseGMLPolygon
					 ( coordinatesString );
			}
			else if ( geometryTypeElementName.getLocalName()
				  .equals( "Box" ) )
			{
				coordinatesString
				  = GeometryFormatConverter
				    .getCoordinatesString( childElement );

				geom = GeometryFormatConverter
					.parseGMLBox
					 ( coordinatesString );
			}
			else
			{
				throw new GeometryException
					( "unexpected geometry type \""
					  + geometryTypeElementName + "\"" );
			}
		}
		catch( GeometryException  e )
		{
			throw new GeometryException( e.getMessage() , e );
		}

		if ( geometryElement.hasAttributeNS( null , "srsName" ) )
		{
			geom.setSrsName( geometryElement.getAttributeNS
					 ( null , "srsName" ) );
		}

		return( geom );
	}

	private	static	String	getCoordinatesString( Element  element )
						      throws GeometryException
	{
		if ( ! element.getNamespaceURI().equals( URI.GML )
		  || ! element.getLocalName().equals( "coordinates" ) )
		{
			throw new GeometryException
				( "unexpected element \""
				  + element.getLocalName()
				  + "\" in namespace "
				  + element.getNamespaceURI()
				  + ", expexted was "
				  + "\"coordinates\" element in namespace "
				  + URI.GML );
		}

		return( ElementUtil.getChildNodesWholeText( element ) );
	}

	private	static	String	getPolygonCoordinatesString( Element  element )
						      throws GeometryException
	{
		Element	coordinatesElement;

		try
		{
			if ( ! element.getNamespaceURI()
			       .equals( URI.GML )
			  || ! element.getLocalName()
			       .equals( "outerBoundaryIs" ) )
			{
				throw new GeometryException
				( "unexpected element \""
				  + element.getLocalName()
				  + "\" in namespace "
				  + element.getNamespaceURI()
				  + ", expexted was "
				  + "\"outerBoundaryIs\" element in namespace "
				  + URI.GML );
			}


			Element	linearRingElement
				= ElementUtil.getSingleChildElement( element );

			System.err.println( "linearRingElement = "
					    + linearRingElement
						    .getLocalName() );

			if ( ! linearRingElement.getNamespaceURI()
			       .equals( URI.GML )
			  || ! linearRingElement.getLocalName()
			       .equals( "LinearRing" ) )
			{
				throw new GeometryException
				( "unexpected element \""
				  + linearRingElement.getLocalName()
				  + "\" in namespace "
				  + linearRingElement.getNamespaceURI()
				  + ", expexted was "
				  + "\"LinearRing\" element in namespace "
				  + URI.GML );
			}

			coordinatesElement
				= ElementUtil.getSingleChildElement
						( linearRingElement );

			System.err.println( "coordinatesElement = "
					    + coordinatesElement
						    .getLocalName() );

		}
		catch( XMLParseErrorException  e )
		{
			throw new GeometryException( e.getMessage() , e );
		}

		if ( ! coordinatesElement.getNamespaceURI()
			.equals( URI.GML )
		  || ! coordinatesElement.getLocalName()
			.equals( "coordinates" ) )
		{
			throw new GeometryException
				( "unexpected element \""
				  + element.getLocalName()
				  + "\" in namespace "
				  + element.getNamespaceURI()
				  + ", expexted was "
				  + "\"coordinates\" element in namespace "
				  + URI.GML );
		}

		return( ElementUtil
			.getChildNodesWholeText( coordinatesElement ) );
	}
}
