package daruma.server;

import daruma.util.PortNumberValidator;
import daruma.sql.DatabaseConnectionException;
import daruma.sql.MySQLDatabaseConnectionFactory;

import java.net.*;
import java.io.*;

public class DarumaDaemon
{
	static	final	public	int	DEFAULT_PORT = 5050;
	static	final	public	String	DEFAULT_DATABASE_NAME = "daruma$devel";
	static	final	public	String	DEFAULT_USER = "daruma";
	static	final	public	String	DEFAULT_PASSWORD = "";

	private	int	port;
	private	String	databaseName;
	private	String	user;
	private	String	password;
	private	boolean	debug;


	public	DarumaDaemon( int  port ,
			      String  databaseName ,
			      String  user ,
			      String  password )
	{
		this.port = port;
		this.debug = false;
		this.databaseName = databaseName;
		this.user = user;
		this.password = password;
	}

	public	boolean	run()
	{
		try
		{
			if ( ! PortNumberValidator.check( port ) )
			{
				System.err.println( "configuration error:"
						    + " invalid port number: "
						    + port );
				return( false );
			}


			ServerSocket	socket = new ServerSocket( port );

			for(;;)
			{
				Socket	s = socket.accept();

				ConnectionHandler	c;

				try
				{
					c = new ConnectionHandler
						  ( s ,
						    this.databaseName ,
						    this.user ,
						    this.password ,
						    this.debug );
				}
				catch( DatabaseConnectionException  e )
				{
					System.err.println( e.getMessage() );
					continue;
				}


				Thread	th = new Thread( c );
				th.start();
			}
		}
		catch( java.net.BindException  e )
		{
			System.err.println( "Can't bind server socket: "
					    + e.getMessage() );
			return( false );
		}
		catch( IOException  e )
		{
			e.printStackTrace();
			return( false );
		}
	}


	public void setDebug( boolean  debug )
	{
		this.debug = debug;
	}


	public static void main( String  args[] )
	{
		try
		{
			DarumaDaemonOptionAnalyzer	opt;
			opt = new DarumaDaemonOptionAnalyzer
				  ( args , "DarumaDaemon" ,
				    DarumaDaemon.DEFAULT_PORT ,
				    DarumaDaemon.DEFAULT_DATABASE_NAME ,
				    DarumaDaemon.DEFAULT_USER ,
				    DarumaDaemon.DEFAULT_PASSWORD );

			if ( ! opt.analyze() )
			{
				opt.printHelp( System.err );

				System.exit( 1 );
			}
			else if ( opt.getHelp() )
			{
				opt.printHelp( System.out );

				System.exit( 0 );
			}


			try
			{
				new MySQLDatabaseConnectionFactory()
							    .loadClass();
			}
			catch( DatabaseConnectionException
			       .ClassLoadFailedDatabaseConnectionException  e )
			{
				System.err.println( "Can't load class: "
						    + e.getFailedClassName() );
				System.exit( 1 );
			}


			DarumaDaemon	daemon = new DarumaDaemon
						     ( opt.getPortNumber() ,
						       opt.getDatabaseName() ,
						       opt.getUser() ,
						       opt.getPassword() );

			daemon.setDebug( opt.getDebug() );

			if ( ! daemon.run() )
			{
				System.exit( 1 );
			}

		}
		catch( Exception  e )
		{
			e.printStackTrace();
			System.exit( 1 );
		}
	}


	public static class  DarumaDaemonOptionAnalyzer
	{
		private	String[]	args;
		private	String		programName;
		private	int		portNumber;
		private	String		databaseName;
		private	String		user;
		private	String		password;

		private	boolean	helpFlag  = false;
		private	boolean	debugFlag = false;

		public	DarumaDaemonOptionAnalyzer
				( final String[]  args ,
				  final String  programName ,
				  final int     defaultPortNumber ,
				  final String  defaultDatabaseName ,
				  final String  defaultUser ,
				  final String  defaultPassword )
		{
			this.args = (String[])args.clone();
			this.programName = programName;
			this.portNumber = defaultPortNumber;
			this.databaseName = defaultDatabaseName;
			this.user = defaultUser;
			this.password = defaultPassword;
		}

		public	boolean	analyze()
		{
			int	index = 0;
			while( index < args.length )
			{
				final	String	arg = this.args[index];
				int	numArgs = args.length - index - 1;

				if ( arg.equals( "--help" ) )
				{
					this.helpFlag = true;
				}
				else if ( arg.equals( "--port" ) )
				{
					if ( numArgs < 1 )
					{
						return( false );
					}

					try
					{
						this.portNumber
						  = Integer.parseInt
						    ( this.args[index + 1] );

					}
					catch( NumberFormatException  e )
					{
						System.err.println
						( "invalid port number: "
						  + this.args[index + 1] );

						return( false );
					}

					index ++;
				}
				else if ( arg.equals( "--database" ) )
				{
					if ( numArgs < 1 )
					{
						return( false );
					}

					this.databaseName
						= this.args[index + 1];

					index ++;
				}
				else if ( arg.equals( "--user" ) )
				{
					if ( numArgs < 1 )
					{
						return( false );
					}

					this.user = this.args[index + 1];

					index ++;
				}
				else if ( arg.equals( "--password" ) )
				{
					if ( numArgs < 1 )
					{
						return( false );
					}

					this.password = this.args[index + 1];

					index ++;
				}
				else if ( arg.equals( "--debug" ) )
				{
					this.debugFlag = true;
				}
				else
				{
					System.err.println
						( "invalid option: " + arg );
					return( false );
				}

				index ++;
			}

			return( true );
		}

		public	void	printHelp( OutputStream  out )
		{
			String	mes = "";
			mes += "Usage: "
				   + this.programName + " [options]" + "\n";
			mes += "Possible options are:"               + "\n";
			mes += "      --help"                        + "\n";
			mes += "      --port PORT"                   + "\n";
			mes += "      --database DATABASE_NAME"      + "\n";
			mes += "      --user USER_NAME"              + "\n";
			mes += "      --password PASSWORD"           + "\n";
			mes += "      --debug"                       + "\n";

			PrintStream	p = new PrintStream( out );
			p.print( mes );
		}

		public	boolean	getHelp()
		{
			return( this.helpFlag );
		}

		public	int	getPortNumber()
		{
			return( this.portNumber );
		}

		public	String	getDatabaseName()
		{
			return( this.databaseName );
		}

		public	String	getUser()
		{
			return( this.user );
		}

		public	String	getPassword()
		{
			return( this.password );
		}

		public	boolean	getDebug()
		{
			return( this.debugFlag );
		}
	}
}
