package daruma.sql;

import daruma.sql.DatabaseConnectionException;

import java.sql.Statement;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.Timestamp;

import java.util.Date;
import java.util.List;
import java.util.ArrayList;


public abstract class DatabaseConnection
{
	abstract protected String	getDatabaseURIPrefix();
	abstract protected String	getDatabaseURIPostfix();


	static	public class QueryResult
	{
		private	java.sql.ResultSet	resultSet;
		private	java.sql.Statement	statement;

		public	QueryResult( java.sql.ResultSet  resultSet ,
				     java.sql.Statement  statement )
		{
			this.resultSet = resultSet;
			this.statement = statement;
		}

		public	java.sql.ResultSet	getResultSet()
					  throws DatabaseConnectionException
		{
			return( this.resultSet );
		}

		public	void	close() throws DatabaseConnectionException
		{
			if ( this.statement != null )
			{
				try
				{
					this.statement.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "statement can't close" ,
						    e );
				}
			}
		}


		public	long	getSingleLongValue( int  columnIndex )
					    throws DatabaseConnectionException
		{
			boolean	found = false;

			long	value = 0L;

			try
			{
				while( this.resultSet.next() )
				{
					if ( found )
					{
						throw
						new DatabaseConnectionException
						  ( "found multiple data"
						    + " in column "
						    + columnIndex );
					}

					found = true;

					value = this.resultSet
						      .getLong( columnIndex );
				}
			}
			catch( java.sql.SQLException  e )
			{
				throw new DatabaseConnectionException( e );
			}

			if ( ! found )
			{
				throw new DatabaseConnectionException
					  ( "data not found in column "
					    + columnIndex );
			}

			return value;
		}
	}

	private	static final int	CONNECTION_MAX_TRIALS = 50;
	private	static final int	CONNECTION_TRIAL_WAIT_MSEC = 1000;

	private	java.sql.Connection	connection;

	DatabaseConnection()
	{
		this.connection = null;
	}

	public	void	connect( String  databaseName ,
				 String  user ,  String  passwd )
					  throws DatabaseConnectionException
	{
		int	n = 1;
		for ( ;  n < CONNECTION_MAX_TRIALS  ;  ++ n )
		{
			try
			{
				String	databaseURI
					= this.getDatabaseURIPrefix()
					  + databaseName;

				String	postfix = this.getDatabaseURIPostfix();

				if ( postfix != null )
				{
					databaseURI += postfix;
				}

				this.connection
				     = java.sql.DriverManager.getConnection
					( databaseURI ,
					  user ,
					  passwd );

				/*
				//
				// do commit/rollback manually
				//
				this.connection.setAutoCommit( false );
				*/

				break;
			}
			catch( Exception  e )
			{
				System.err.println();
				System.err.println
					( e.getMessage()
					  + "recconect count = " + n );

				try
				{
					Thread.sleep
						( CONNECTION_TRIAL_WAIT_MSEC );
				}
				catch( InterruptedException  ie )
				{
					ie.printStackTrace();
				}
			}
		}

		if ( n > CONNECTION_MAX_TRIALS )
		{
			throw new DatabaseConnectionException
				  ( "connection not established" );
		}

		if ( n != 1 )
		{
			System.err.println( "connected!!" );
		}
	}


	public	void	close() throws DatabaseConnectionException
	{
		if ( this.connection != null )
		{
			try
			{
				this.connection.close();
			}
			catch( java.sql.SQLException  e )
			{
				throw new DatabaseConnectionException
				( "connection closing failed,"
				  + " sql error ["
				  + e.getMessage() + "]" , e );
			}
		}
	}

	public void	startTransaction() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "start transaction failed,"
			      + " connection not established" );
		}

		try
		{
			this.executeQuery( "START TRANSACTION" );
		}
		catch( DatabaseConnectionException e )
		{
			throw new DatabaseConnectionException
				  ( "start transaction failed: "
				    + e.getMessage() );
		}
	}

	public void	endTransaction() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "end transaction failed,"
			      + " connection not established" );
		}
	}

	public void	commit() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "commit transaction failed,"
			      + " connection not established" );
		}

		try
		{
			this.executeQuery( "COMMIT" );
		}
		catch( DatabaseConnectionException e )
		{
			throw new DatabaseConnectionException
				  ( "commit failed: " + e.getMessage() );
		}
	}

	public void	rollback() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "rollback failed,"
			      + " connection not established" );
		}

		try
		{
			this.executeQuery( "ROLLBACK" );
		}
		catch( DatabaseConnectionException e )
		{
			throw new DatabaseConnectionException
				  ( "commit failed: " + e.getMessage() );
		}
	}


	public	java.util.Date	getCurrentTime()
					    throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
				  ( "can't get current time,"
				    + " connection is not established yet" );
		}


		Statement	s = null;

		try
		{
			s = this.connection.createStatement();

			ResultSet	r = s.executeQuery
						( "SELECT NOW() time" );


			java.sql.Timestamp	ret = null;

			if( r.next() )
			{
				ret = r.getTimestamp( "time" );
			}
			else
			{
				throw new DatabaseConnectionException
				  ( "can't get current time,"
				    + " unexpected database response" );
			}

			if ( r.next() )
			{
				throw new DatabaseConnectionException
				  ( "can't get current time,"
				    + " unexpected database response" );
			}

			return( ret );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
			  ( "can't get current time,"
			    + " sql error [" + e.getMessage() + "]" , e );
		}
		finally
		{
			if ( s != null )
			{
				try
				{
					s.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
					  ( "connection closing failed,"
					    + " sql error ["
					    + e.getMessage() + "]" , e );
				}
			}
		}
	}


	public	PreparedStatement	prepareStatement( String  sqlString )
					  throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "prepared statement failed,"
			      + " connection not established" );
		}

		try
		{
			return this.connection.prepareStatement( sqlString );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "prepaerd statement failed: "
				    + e.getMessage() , e );
		}
	}



	public	QueryResult	executeQuery( String  sqlString )
					  throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "query failed, connection not established" );
		}

		Statement	s = null;
		try
		{
			System.err.println( "#" + sqlString );

			s = this.connection.createStatement();

			ResultSet	r = s.executeQuery( sqlString );

			return( new QueryResult( r , s ) );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "query failed: " + e.getMessage() , e );
		}
	}

	public	QueryResult	executeQuery( PreparedStatement  statement )
					  throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "query failed, connection not established" );
		}

		try
		{
			ResultSet	r = statement.executeQuery();

			return( new QueryResult( r , statement ) );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "query failed: " + e.getMessage() , e );
		}
	}


	public	int	executeUpdate( String  sqlString )
					  throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
				  ( "connection not established" );
		}

		Statement	s = null;
		try
		{
			s = this.connection.createStatement();

			System.err.println( "#" + sqlString );

			int	count = s.executeUpdate( sqlString );

			return( count );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "update failed: " + e.getMessage() , e );
		}
		finally
		{
			if ( s != null )
			{
				try
				{
					s.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "statement can't close" ,
						    e );
				}
			}
		}
	}

	public	int	executeUpdate( PreparedStatement  statement )
					  throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
				  ( "connection not established" );
		}

		try
		{
			int	count = statement.executeUpdate();

			return( count );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "update failed: " + e.getMessage() , e );
		}
		finally
		{
			if ( statement != null )
			{
				try
				{
					statement.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "statement can't close" ,
						    e );
				}
			}
		}
	}


	/*
	// XXX: should handle table column definition
	public	void	createTableIfNotExists( String  tableName )
					    throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "can't create table [" + tableName + "]"
			      + " because connection is not established yet" );
		}

		Statement	s = null;
		try
		{
			s = this.connection.createStatement();

			// XXX: "IF NOT EXISTS" is MySQL extension

			// XXX
			int	r = s.executeUpdate( "CREATE TABLE"
						     + " IF NOT EXISTS "
						     + tableName
						     + " (abc INTEGER)" );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
			    ( "create table [" + tableName + "] failed" , e );
		}
		finally
		{
			if ( s != null )
			{
				try
				{
					s.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "statement can't close" ,
						    e );
				}
			}
		}
	}
	*/

	public	void	dropAllTables() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "can't drop all tables"
			      + " because connection is not established yet" );
		}

		Statement	s = null;
		ResultSet	r = null;

		try
		{
			s = this.connection.createStatement();
			r = s.executeQuery( "SHOW TABLES" );

			while( r.next() )
			{
				Statement	d = null;

				try
				{
					d = this.connection.createStatement();

					String	sql = "DROP TABLE"
						      + " IF EXISTS `"
						      + r.getString( 1 )
						      + "`";

					System.err.println( "#" + sql );

					d.executeUpdate( "DROP TABLE"
							 + " IF EXISTS `"
							 + r.getString( 1 )
							 + "`" );
					d.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						( "`drop table' failed" , e );
				}
				finally
				{
					if ( d != null )
					{
						try
						{
							d.close();
						}
						catch( java.sql.SQLException  e )
						{
							throw new DatabaseConnectionException
								( "statement can't close" ,
								  e );
						}
					}
				}
			}
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "`drop all tables' failed" , e );
		}
		finally
		{
			if ( s != null )
			{
				try
				{
					s.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "statement can't close" ,
						    e );
				}
			}
		}
	}


	public	void	dropTableIfExists( String  tableName )
					   throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "can't drop table [" + tableName + "]"
			      + " because connection is not established yet" );
		}


		Statement	d = null;

		try
		{
			d = this.connection.createStatement();

			// XXX: "IF EXISTS" is MySQL extention,
			//      should move to MySQLDatabaseConnection

			String	sql = "DROP TABLE IF EXISTS "
				      + "`" + tableName + "`";
			System.err.println( "#" + sql );

			d.executeUpdate( "DROP TABLE IF EXISTS "
					 + "`" + tableName + "`" );
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				( "`drop table' failed" , e );
		}
		finally
		{
			if ( d != null )
			{
				try
				{
					d.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						( "statement can't close" ,
						  e );
				}
			}
		}
	}


	public	List<String>	showTables() throws DatabaseConnectionException
	{
		if ( this.connection == null )
		{
			throw new DatabaseConnectionException
			    ( "can't show tables"
			      + " because connection is not established yet" );
		}


		List<String>	ret = new ArrayList<String>();

		Statement	s = null;
		try
		{
			s = this.connection.createStatement();

			ResultSet	r = s.executeQuery( "SHOW TABLES" );

			while( r.next() )
			{
				ret.add( r.getString( /*column*/ 1 ) );
			}
		}
		catch( java.sql.SQLException  e )
		{
			throw new DatabaseConnectionException
				  ( "show tables failed" , e );
		}
		finally
		{
			if ( s != null )
			{
				try
				{
					s.close();
				}
				catch( java.sql.SQLException  e )
				{
					throw new DatabaseConnectionException
						  ( "can't close statement" ,
						    e );
				}
			}
		}

		return( ret );
	}


	public	long	getMaxLongValueFromTable( String  tableName ,
						  String  columnName )
					throws DatabaseConnectionException
	{
		QueryResult	r = null;

		try
		{
			r = this.executeQuery
				( "SELECT MAX(`" + columnName + "`)"
				  + " FROM `" + tableName + "`" );

			return( r.getSingleLongValue( 1 ) );
		}
		finally
		{
			if ( r != null )
			{
				r.close();
			}
		}
	}
}
