package daruma.strage_manager.schema;

import daruma.strage_manager.StrageManager;
import daruma.strage_manager.StrageException;

import daruma.strage_manager.type_definition.TypeName;
import daruma.strage_manager.type_definition.TypeDefinition;
import daruma.strage_manager.type_definition.ElementName;

import daruma.strage_manager.type_definition.types.SequenceTypeDefinition;
import daruma.strage_manager.type_definition.types.AllTypeDefinition;
import daruma.strage_manager.type_definition.types.ChoiceTypeDefinition;
import daruma.strage_manager.type_definition.types.StringTypeDefinition;
import daruma.strage_manager.type_definition.types.MultipleOccursElementTypeDefinition;

import daruma.xml.QName;
import daruma.xml.URI;
import daruma.xml.util.PrefixMap;
import daruma.xml.util.ElementUtil;

import daruma.util.Pair;

import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import java.util.Map;
import java.util.List;


public class XMLSchemaTypeParser
{
	private	XMLSchemaTypeParser()
	{
	}

	public	static	TypeDefinition	parse( Element  e ,
					       String  targetNamespace ,
					       PrefixMap  prefixMap ,
					       StrageManager  strage )
				throws XMLSchemaParserException
	{
		if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
		{
			throw new XMLSchemaParserException
				  ( "element \"" + e.getLocalName()
				    + "\" isn't a member of "
				    + URI.XML_SCHEMA );
		}


		if ( e.getLocalName().equals( "simpleType" ) )
		{
			return( parseSimpleTypeOrContent
				( e , targetNamespace ,
				  prefixMap , strage ) );
		}
		else if ( e.getLocalName().equals( "complexType" ) )
		{
			return( parseComplexTypeOrComplexContentExtension
				( e , true , targetNamespace ,
				  prefixMap , strage ) );
		}
		else
		{
			throw new XMLSchemaParserException
				  ( "unexpected element \""
				    + e.getLocalName() + "\"" );
		}
	}


	public	static	TypeDefinition	parseSimpleTypeOrContent
					( Element  simpleTypeElement ,
					  String  targetNamespace ,
					  PrefixMap  prefixMap ,
					  StrageManager  strage )
						throws XMLSchemaParserException
	{
		TypeDefinition	ret = null;

		List<Element>	childs = ElementUtil.getChildElements
							( simpleTypeElement );

		Element	child = null;

		for ( Element  e  :  childs )
		{
			if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
					( "element \""
					  + e.getLocalName()
					  + "\" isn't a member of "
					  + URI.XML_SCHEMA );
			}
			else if ( e.getLocalName().equals( "restriction" ) )
			{
				if ( child != null )
				{
					throw new XMLSchemaParserException
					  ( "too many childs found in \""
					    + simpleTypeElement.getLocalName()
					    + "\"" );
				}

				child = e;
			}
			else if ( e.getLocalName().equals( "annotation" ) )
			{
				XMLSchemaTypeParser.checkAnnotation( e );
			}
			else
			{
				throw new XMLSchemaParserException
					( e.getLocalName()
					  + " in namespace "
					  + e.getNamespaceURI()
					  + " is not a valid child of "
					  + simpleTypeElement.getLocalName() );
			}
		}

		if ( child == null )
		{
			throw new XMLSchemaParserException
				( "unexpected empty child of \""
				  + simpleTypeElement.getLocalName() );
		}

		if ( ! child.hasAttribute( "base" ) )
		{
			throw new XMLSchemaParserException
				  ( child.getLocalName()
				    + " in namespace "
				    + child.getNamespaceURI()
				    + " must have \"base\" attribute" );
		}

		String	baseTypeQName = child.getAttribute( "base" );

		Pair<TypeName, TypeDefinition>	type;

		try
		{
			type = XMLSchemaTypeParser
				.getTypeDefinitionFromQNameString
				  ( baseTypeQName , prefixMap , strage );
		}
		catch( XMLSchemaParserException  e )
		{
			throw e;
		}

		TypeDefinition	typeDefinition = type.getSecond();

		if ( typeDefinition.isXMLComplexType() )
		{
			throw new XMLSchemaParserException
				  ( "base type " + baseTypeQName
				    + " is not a simpleType" );
		}

		return( XMLSchemaTypeParser
			.appendSimpleTypeRestrictionToTypeDefinition
			( typeDefinition , child ) );
	}


	public	static	TypeDefinition
			    appendSimpleTypeRestrictionToTypeDefinition
						( TypeDefinition  type ,
						  Element  restrictonElement )
						throws XMLSchemaParserException
	{
		// XXX: loosely parsing

		TypeDefinition	result = type;

		for ( Element  r
			:  ElementUtil.getChildElements( restrictonElement ) )
		{
			if ( ! r.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
				  ( r.getLocalName()
				    + " in namespace "
				    + r.getNamespaceURI()
				    + " is not a schema element" );
			}

			if ( r.getLocalName().equals( "maxLength" )
			  || r.getLocalName().equals( "minLength" ) )
			{
				if ( ! r.hasAttribute( "value" ) )
				{
					throw new XMLSchemaParserException
					( r.getLocalName()
					  + " in namespace "
					  + r.getNamespaceURI()
					  + " does not have attribute"
					  + " \"value\"" );
				}

				String	value = r.getAttribute( "value" );

				long	limit;

				try
				{
					limit = Long.decode( value );
				}
				catch( NumberFormatException  e )
				{
					throw new XMLSchemaParserException
					( "\"" + value + "\""
					  + " is not a valid integer" );
				}

				if ( limit < 0 )
				{
					throw new XMLSchemaParserException
					( "\"" + value + "\""
					  + " is not a unsigned integer" );
				}

				// XXX
				if ( type instanceof StringTypeDefinition )
				{
					result = ((StringTypeDefinition)result)
								.deepCopy();

					if ( r.getLocalName().equals
					     ( "minLength" ) )
					{
						((StringTypeDefinition)result)
							.setMinLength( limit );
					}
					else if ( r.getLocalName().equals
						  ( "maxLength" ) )
					{
						((StringTypeDefinition)result)
							.setMaxLength( limit );
					}
				}
			}
			else if ( r.getLocalName().equals( "enumeration" ) )
			{
				if ( ! r.hasAttribute( "value" ) )
				{
					throw new XMLSchemaParserException
					( r.getLocalName()
					  + " in namespace "
					  + r.getNamespaceURI()
					  + " does not have attribute"
					  + " \"value\"" );
				}

				String	value = r.getAttribute( "value" );

				// XXX
				if ( type instanceof StringTypeDefinition )
				{
					result = ((StringTypeDefinition)result)
								.deepCopy();

					((StringTypeDefinition)result)
						.addEnumeration( value );
				}
			}
			else if ( r.getLocalName().equals( "annotation" ) )
			{
				XMLSchemaTypeParser.checkAnnotation( r );
			}
			else
			{
				// XXX: unsupported yet
			}
		}

		return( result );
	}


	public	static	TypeDefinition
			  parseComplexTypeOrComplexContentExtension
					( Element  complexTypeElement ,
					  boolean  allowExtraAttributes ,
					  String  targetNamespace ,
					  PrefixMap  prefixMap ,
					  StrageManager  strage )
						throws XMLSchemaParserException
	{
		TypeDefinition	ret = null;

		NodeList	childs = complexTypeElement.getChildNodes();

		int	i = 0;

		for ( ;  i < childs.getLength()  ;  ++ i )
		{
			Node	n = childs.item(i);

			if ( ! (n instanceof Element) )
			{
				continue;
			}

			Element	e = (Element)n;

			if ( e.getLocalName().equals( "sequence" )
			  && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				ret = XMLSchemaTypeParser
				      .parseSequence( e , targetNamespace ,
						      prefixMap , strage ,
						      true );
			}
			else if ( e.getLocalName().equals( "all" )
			       && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				ret = XMLSchemaTypeParser
				      .parseAll( e , targetNamespace ,
						 prefixMap , strage ,
						 true );
			}
			else if ( e.getLocalName().equals( "simpleContent" )
			       && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				ret = XMLSchemaTypeParser
					.parseSimpleTypeOrContent
					 ( e , targetNamespace ,
					   prefixMap , strage );
			}
			else if ( e.getLocalName().equals( "complexContent" )
			       && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				ret = XMLSchemaTypeParser
					.parseComplexContent
					 ( e , targetNamespace ,
					   prefixMap , strage );
			}
			else if ( e.getLocalName().equals( "choice" )
			       && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				ret = XMLSchemaTypeParser
					.parseChoice
					 ( e , targetNamespace ,
					   prefixMap , strage ,
					   false );
			}
			else if ( e.getLocalName().equals( "annotation" )
			       && e.getNamespaceURI().equals(URI.XML_SCHEMA) )
			{
				// check only
				XMLSchemaTypeParser.checkAnnotation( e );

				continue;
			}
			else
			{
				throw new XMLSchemaParserException
					  ( "unexpected element \""
					    + e.getLocalName()
					    + "\" in namespace "
					    + e.getNamespaceURI() );
			}

			i++;
			break;
		}

		if ( allowExtraAttributes )
		{
			for ( ;  i < childs.getLength()  ;  ++ i )
			{
				Node	n = childs.item(i);

				if ( ! (n instanceof Element) )
				{
					continue;
				}

				Element	e = (Element)n;

				if ( e.getLocalName().equals( "attribute" )
				  && e.getNamespaceURI()
					     .equals(URI.XML_SCHEMA) )
				{
					// XXX: not supported yet
				}
				else if ( e.getLocalName()
						.equals( "annotation" )
				       && e.getNamespaceURI()
						.equals(URI.XML_SCHEMA) )
				{
					// check only
					XMLSchemaTypeParser
						.checkAnnotation( e );
				}
				else
				{
					throw new XMLSchemaParserException
						( "unexpected element \""
						  + e.getLocalName()
						  + "\" in namespace "
						  + e.getNamespaceURI() );
				}
			}
		}

		if ( ret != null )
		{
			return( ret );
		}

		throw new XMLSchemaParserException
		  ( "too few childs in \""
		    + complexTypeElement.getLocalName()
		    + "\"" );
	}


	public	static	TypeDefinition	parseComplexContent
					( Element  complexContentElement ,
					  String  targetNamespace ,
					  PrefixMap  prefixMap ,
					  StrageManager  strage )
						throws XMLSchemaParserException
	{
		List<Element>	childs = ElementUtil.getChildElements
						     ( complexContentElement );

		Element	child = null;

		for ( Element  e  :  childs )
		{
			if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
					( "element \""
					  + e.getLocalName()
					  + "\" isn't a member of "
					  + URI.XML_SCHEMA );
			}
			else if ( e.getLocalName().equals( "annotation" ) )
			{
				XMLSchemaTypeParser.checkAnnotation( e );
			}
			else
			{
				if ( child != null )
				{
					throw new XMLSchemaParserException
					  ( "too many childs found in \""
					    + complexContentElement
						    .getLocalName() + "\"" );
				}

				child = e;
			}
		}

		if ( child == null )
		{
			throw new XMLSchemaParserException
				( "unexpected empty child of \""
				  + complexContentElement.getLocalName() );
		}


		if ( child.getLocalName().equals( "extension" ) )
		{
		}
		else if ( child.getLocalName().equals( "restriction" ) )
		{
			// XXX: not implemented yet
			throw new XMLSchemaParserException
			  ( child.getLocalName()
			    + " in namespace "
			    + child.getNamespaceURI()
			    + " in complexType has not been implemented yet" );
		}
		else
		{
			throw new XMLSchemaParserException
				  ( child.getLocalName()
				    + " in namespace "
				    + child.getNamespaceURI()
				    + " is not a valid child of "
				    + complexContentElement.getLocalName() );
		}


		//
		// child is `extension`
		//
		if ( ! child.hasAttribute( "base" ) )
		{
			throw new XMLSchemaParserException
				  ( child.getLocalName()
				    + " in namespace "
				    + child.getNamespaceURI()
				    + " must have \"base\" attribute" );
		}

		String	baseTypeQName = child.getAttribute( "base" );

		Pair<TypeName, TypeDefinition>	type;

		try
		{
			type = XMLSchemaTypeParser
				.getTypeDefinitionFromQNameString
				  ( baseTypeQName , prefixMap , strage );
		}
		catch( XMLSchemaParserException  e )
		{
			throw e;
		}

		TypeDefinition	baseTypeDefinition = type.getSecond();

		if ( ! baseTypeDefinition.isXMLComplexType() )
		{
			throw new XMLSchemaParserException
				  ( "base type" + baseTypeQName
				    + " is not a complexType" );
		}


		TypeDefinition	additionalDefinition
				= XMLSchemaTypeParser
				  .parseComplexTypeOrComplexContentExtension
							    ( child ,
							      false ,
							      targetNamespace ,
							      prefixMap ,
							      strage );

		SequenceTypeDefinition	seq = new SequenceTypeDefinition();

		seq.addChild( baseTypeDefinition );
		seq.addChild( additionalDefinition );

		return( seq );
	}


	public	static	SequenceTypeDefinition
				parseSequence( Element  sequenceElement ,
					       String  targetNamespace ,
					       PrefixMap  prefixMap ,
					       StrageManager  strage ,
					       boolean  isTopLevel )
						throws XMLSchemaParserException
	{
		SequenceTypeDefinition	seq = new SequenceTypeDefinition();

		NodeList	childs = sequenceElement.getChildNodes();

		for ( int  i = 0  ;  i < childs.getLength()  ;  ++ i )
		{
			Node	n = childs.item(i);

			if ( ! (n instanceof Element) )
			{
				continue;
			}

			Element	e = (Element)n;

			if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
					  ( "invalid tag: " + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + ": expected namespace was "
					    + URI.XML_SCHEMA );
			}

			if ( e.getLocalName().equals( "element" ) )
			{
				if ( ! e.hasAttribute( "name" )
				  || ! e.hasAttribute( "type" ) )
				{
					// XXX: tentative specification
					throw new XMLSchemaParserException
					  ( e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + " must have"
					    + " \"name\" and \"type\""
					    + " attribute" );
				}


				String	name = e.getAttribute( "name" );
				String	typeQNameString = e.getAttribute
							       ( "type" );
				QName	typeQName = new QName
							  ( typeQNameString );

				if ( ! typeQName.isValid() )
				{
					throw new XMLSchemaParserException
					  ( "type \""
					    + e.getAttribute( "type" )
					    + "\" is not valid QName" );
				}


				String	typeNamespace
					= prefixMap
					  .get( typeQName.getPrefix() );

				if ( typeNamespace == null )
				{
					throw new XMLSchemaParserException
					  ( "prefix of \""
					    + typeQNameString
					    + "\" was not found" );
				}


				// XXX
				System.err.println( "element name = ["
						    + name + "]" );
				System.err.println( "element type = [{"
						    + typeNamespace
						    + "}"
						    + typeQName.getLocalName()
						    + "]" );

				TypeName  typeName
					  = new TypeName
						( typeNamespace ,
						  typeQName.getLocalName() );

				TypeDefinition	type;
				try
				{
					type = strage
						.getTypeDefinition( typeName );
				}
				catch( StrageException  se )
				{
					throw new XMLSchemaParserException
									( se );
				}


				if ( type == null )
				{
					throw new XMLSchemaParserException
					  ( "type \""
					    + typeName.getLocalName()
					    + "\" in namespace \""
					    + typeName.getNamespace()
					    + "\" not found" );
				}

				System.err.println
					  ( "type = ["
					    + typeName.getLocalName()
					    + "], namespace = ["
					    + typeName.getNamespace()
					    + "]" );

				ElementName	elementName
						= new ElementName
						  ( targetNamespace ,
						    name );

				try
				{
					strage
					 .registerUniversalNameIfNotRegistrated
							       ( elementName );
				}
				catch( StrageException  se )
				{
					throw new XMLSchemaParserException
					  ( "can't register " + elementName ,
					    se );
				}


				int	minOccurs = 1;
				int	maxOccurs = 1;

				if ( e.hasAttribute( "maxOccurs" ) )
				{
					try
					{
						maxOccurs
						  = Integer.parseInt
						    ( e.getAttribute
						      ( "maxOccurs" ) );
					}
					catch( NumberFormatException ne )
					{
						throw new XMLSchemaParserException
						 ( "value of maxOccurs"
						   + " not a valid integer" );
					}
				}

				if ( e.hasAttribute( "minOccurs" ) )
				{
					try
					{
						minOccurs
						  = Integer.parseInt
						    ( e.getAttribute
						      ( "minOccurs" ) );
					}
					catch( NumberFormatException ne )
					{
						throw new XMLSchemaParserException
						 ( "value of minOccurs"
						   + " not a valid integer" );
					}
				}

				if ( minOccurs < 0 )
				{
					throw new XMLSchemaParserException
					   ( "negative value of minOccurs" );
				}
				else if ( minOccurs < 0 )
				{
					throw new XMLSchemaParserException
					   ( "negative value of minOccurs" );
				}
				else if ( minOccurs > maxOccurs )
				{
					throw new XMLSchemaParserException
					 ( "minOccurs should be"
					   + " less or equals to maxOccurs" );
				}

				if ( maxOccurs >= 2 )
				{
					seq.addChild
					( new MultipleOccursElementTypeDefinition
					  ( typeName , elementName ,
					    minOccurs , maxOccurs ) );
				}
				else if ( minOccurs == 0 )
				{
				}
				else
				{
					seq.addChild( typeName , elementName );
				}
			}
			else if ( e.getLocalName().equals( "sequence" ) )
			{
				seq.addChild( parseSequence
					      ( e ,
						targetNamespace ,
						prefixMap ,
						strage ,
						false ) );
			}
			else if ( e.getLocalName().equals( "choice" ) )
			{
				seq.addChild( parseChoice
					      ( e ,
						targetNamespace ,
						prefixMap ,
						strage ,
						false ) );
			}
			else if ( e.getLocalName().equals( "annotation" ) )
			{
				// check only
				XMLSchemaTypeParser.checkAnnotation( e );
			}
			else
			{
				throw new XMLSchemaParserException
					  ( "unexpected type: "
					    + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI() );
			}
		}

		return( seq );
	}


	public	static	AllTypeDefinition
				parseAll( Element  allElement ,
					  String  targetNamespace ,
					  PrefixMap  prefixMap ,
					  StrageManager  strage ,
					  boolean  isTopLevel )
						throws XMLSchemaParserException
	{
		AllTypeDefinition	all = new AllTypeDefinition();

		ElementName	allElementName = new ElementName( allElement );

		NodeList	childs = allElement.getChildNodes();

		for ( int  i = 0  ;  i < childs.getLength()  ;  ++ i )
		{
			Node	n = childs.item(i);

			if ( ! (n instanceof Element) )
			{
				continue;
			}

			Element	e = (Element)n;

			if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
					  ( "invalid tag: " + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + ": expected namespace was "
					    + URI.XML_SCHEMA );
			}

			if ( e.getLocalName().equals( "element" ) )
			{
				if ( ! e.hasAttribute( "name" )
				  || ! e.hasAttribute( "type" ) )
				{
					// XXX: tentative specification
					throw new XMLSchemaParserException
					  ( e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + "must have"
					    + " \"type\" attribute" );
				}


				String	name = e.getAttribute( "name" );

				Pair<TypeName, TypeDefinition>	type;

				try
				{
					type
					  = XMLSchemaTypeParser
					     .getTypeDefinitionFromQNameString
					      ( e.getAttribute( "type" ) ,
						prefixMap ,
						strage );
				}
				catch( XMLSchemaParserException  pe )
				{
					throw pe;
				}

				ElementName	elementName
						= new ElementName
						  ( targetNamespace ,
						    name );

				try
				{
					strage
					 .registerUniversalNameIfNotRegistrated
							       ( elementName );
				}
				catch( StrageException  se )
				{
					throw new XMLSchemaParserException
					  ( "can't register " + elementName ,
					    se );
				}

				all.addChild( type.getFirst() , elementName );

			}
			else if ( e.getLocalName().equals( "annotation" ) )
			{
				// check only
				XMLSchemaTypeParser.checkAnnotation( e );
			}
			else if ( e.getLocalName().equals( "sequence" ) )
			{
				throw new XMLSchemaParserException
					  ( "unexpected element \""
					    + e.getLocalName()
					    + "\" in namespace "
					    + e.getNamespaceURI()
					    + " as a child of "
					    + allElementName.toString() );
			}
			else
			{
				throw new XMLSchemaParserException
					  ( "unexpected type: "
					    + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + ", expected "
					    + "element or annotation" );
			}
		}

		return( all );
	}


	public	static	ChoiceTypeDefinition
				parseChoice( Element  choiceElement ,
					     String  targetNamespace ,
					     PrefixMap  prefixMap ,
					     StrageManager  strage ,
					     boolean  isTopLevel )
						throws XMLSchemaParserException
	{
		ChoiceTypeDefinition	choice = new ChoiceTypeDefinition();

		NodeList	childs = choiceElement.getChildNodes();

		for ( int  i = 0  ;  i < childs.getLength()  ;  ++ i )
		{
			Node	n = childs.item(i);

			if ( ! (n instanceof Element) )
			{
				continue;
			}

			Element	e = (Element)n;

			if ( ! e.getNamespaceURI().equals( URI.XML_SCHEMA ) )
			{
				throw new XMLSchemaParserException
					  ( "invalid tag: " + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + ": expected namespace was "
					    + URI.XML_SCHEMA );
			}

			if ( e.getLocalName().equals( "element" ) )
			{
				// XXX: this code should be moved to a method

				if ( ! e.hasAttribute( "name" )
				  || ! e.hasAttribute( "type" ) )
				{
					// XXX: tentative specification
					throw new XMLSchemaParserException
					  ( e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + "must have"
					    + " \"type\" attribute" );
				}


				String	name = e.getAttribute( "name" );

				Pair<TypeName, TypeDefinition>	type;

				try
				{
					type
					  = XMLSchemaTypeParser
					     .getTypeDefinitionFromQNameString
					      ( e.getAttribute( "type" ) ,
						prefixMap ,
						strage );
				}
				catch( XMLSchemaParserException  pe )
				{
					throw pe;
				}

				ElementName	elementName
						= new ElementName
						  ( targetNamespace ,
						    name );

				try
				{
					strage
					 .registerUniversalNameIfNotRegistrated
							       ( elementName );
				}
				catch( StrageException  se )
				{
					throw new XMLSchemaParserException
					  ( "can't register " + elementName ,
					    se );
				}

				choice.addChild( type.getFirst() ,
						 elementName );
			}
			else if ( e.getLocalName().equals( "sequence" ) )
			{
				choice.addChild( XMLSchemaTypeParser
						 .parseSequence
						   ( e , targetNamespace ,
						     prefixMap , strage ,
						     false ) );
			}
			else if ( e.getLocalName().equals( "choice" ) )
			{
				choice.addChild( XMLSchemaTypeParser
						 .parseChoice
						   ( e , targetNamespace ,
						     prefixMap , strage ,
						     false ) );
			}
			else if ( e.getLocalName().equals( "group" )
			       || e.getLocalName().equals( "any" ) )
			{
				// XXX: not implemented yet
				throw new XMLSchemaParserException
				  ( new ElementName( choiceElement ).toString()
				    + " in " + choiceElement
				    + " has not been implemented yet" );
			}
			else if ( e.getLocalName().equals( "annotation" ) )
			{
				//
				// XXX: should check this element is
				//      first child or not

				// check only
				XMLSchemaTypeParser.checkAnnotation( e );
			}
			else
			{
				throw new XMLSchemaParserException
					  ( "unexpected type: "
					    + e.getLocalName()
					    + " in namespace "
					    + e.getNamespaceURI()
					    + ", expected "
					    + "element or annotation" );
			}
		}

		return( choice );
	}



	//
	// utilities
	//
	static	Pair<TypeName, TypeDefinition>
		getTypeDefinitionFromQNameString( String  typeQNameString ,
						  PrefixMap  prefixMap ,
						  StrageManager  strage )
						throws XMLSchemaParserException
	{
		QName	typeQName = new QName( typeQNameString );

		if ( ! typeQName.isValid() )
		{
			throw new XMLSchemaParserException
			  ( "type \"" + typeQNameString
			    + "\" is not a valid QName" );
		}

		String	typeNamespace = prefixMap.get( typeQName.getPrefix() );

		if ( typeNamespace == null )
		{
			throw new XMLSchemaParserException
				  ( "prefix \"" + typeQName.getPrefix()
				    + "\" was not found" );
		}

		// XXX
		System.err.println( "element type = [{" + typeNamespace + "}"
				    + typeQName.getLocalName() + "]" );

		TypeName  typeName = new TypeName( typeNamespace ,
						   typeQName.getLocalName() );

		TypeDefinition	type;
		try
		{
			type = strage.getTypeDefinition( typeName );
		}
		catch( StrageException  se )
		{
			throw new XMLSchemaParserException( se );
		}


		if ( type == null )
		{
			throw new XMLSchemaParserException
				( "type \"" + typeName.getLocalName()
				  + "\" in namespace \""
				  + typeName.getNamespace() + "\" not found" );
		}

		// XXX
		System.err.println( "type = ["
				    + typeName.getLocalName()
				    + "], namespace = ["
				    + typeName.getNamespace()
				    + "]" );

		return( new Pair<TypeName, TypeDefinition>
			( typeName , type ) );
	}


	public	static	void	checkAnnotation( Element  element )
					throws XMLSchemaParserException
	{
		// XXX: no checks
	}
}
